window.Gleam.initSearch([{"doc":"ream","title":"ream","content":"# ream\n\n[![Package Version](https://img.shields.io/hexpm/v/ream)](https://hex.pm/packages/ream)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/ream/)\n\nREAM is an event-sourcing system.\n\nThe REAM server is a simple TCP server that&#39;s receiving events from aggregators which are\nsending the event and a status update. The idea is:\n\n- The Aggregator receives a command.\n- The aggregator validates the command and generates:\n    1. One or more events are generated by the aggregator and published into REAM.\n    2. A modification in the state is performed in the aggregator and stored in REAM.\n- Subscribers, mainly projectors receive the event.\n- These subscribers perform a change in the schema and publish changes into projections.\n\nTherefore we have three different but related storage:\n\n- Events. It&#39;s getting events in different streams. We can create as many streams as we need.\n- Aggregations. It&#39;s storing the state for the aggregators based on the last processed event for a specific stream.\n- Projections. Stored the projections based on the information provided by the projectors.\n\nThe main idea is to provide an easy way for implementing aggregators and projectors to process the data we need in the way we need it.\n\n## Quick start\n\n```sh\ngleam run   # Run the project\ngleam test  # Run the tests\ngleam shell # Run an Erlang shell\n```\n\n## Installation\n\nIf available on Hex this package can be added to your Gleam project:\n\n```sh\ngleam add ream\n```\n\nand its documentation can be found at &lt;https://hexdocs.pm/ream&gt;.\n\n## Architecture and Design\n\nYou can check the protocol [here](docs/protocol.md) and how the storage is going to take place [here](docs/storage.md).\n\n## License\n\nApache License Version 2.0, see [LICENSE](LICENSE)\n\nCopyright 2023 Altenwald\n\nLicensed under the Apache License, Version 2.0 (the &quot;License&quot;);\nyou may not use this file except in compliance with the License.\nYou may obtain a copy of the License at\n\nhttp://www.apache.org/licenses/LICENSE-2.0\n\nUnless required by applicable law or agreed to in writing, software\ndistributed under the License is distributed on an &quot;AS IS&quot; BASIS,\nWITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\nSee the License for the specific language governing permissions and\nlimitations under the License.\n","url":"index.html"},{"doc":"ream","title":"ream","content":"                                 Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      &quot;License&quot; shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      &quot;Licensor&quot; shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      &quot;Legal Entity&quot; shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      &quot;control&quot; means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      &quot;Source&quot; form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      &quot;Object&quot; form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      &quot;Work&quot; shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      &quot;Derivative Works&quot; shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      &quot;Contribution&quot; shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, &quot;submitted&quot;\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as &quot;Not a Contribution.&quot;\n\n      &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a &quot;NOTICE&quot; text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS\n\n   APPENDIX: How to apply the Apache License to your work.\n\n      To apply the Apache License to your work, attach the following\n      boilerplate notice, with the fields enclosed by brackets &quot;[]&quot;\n      replaced with your own identifying information. (Don&#39;t include\n      the brackets!)  The text should be enclosed in the appropriate\n      comment syntax for the file format. We also recommend that a\n      file or class name and description of purpose be included on the\n      same &quot;printed page&quot; as the copyright notice for easier\n      identification within third-party archives.\n\n   Copyright [yyyy] [name of copyright owner]\n\n   Licensed under the Apache License, Version 2.0 (the &quot;License&quot;);\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an &quot;AS IS&quot; BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.\n","url":"license.html"},{"doc":"ream","title":"ream","content":"# Protocol\n\nThe protocol to be connected to the database is based on plain TCP and no credentials for the first version. The system is based on asynchronous communication, when we are connected the system is giving us the banner:\n\n```\nREAM 1.0\n```\n\nAfter that, we can send whenever a command and the server is sending us whenever information. For example, if an event is published by another system, we are receiving this:\n\n```\nEVENT emails 1001 16\n{&quot;name&quot;:&quot;peter&quot;}\n```\n\nThat&#39;s telling us that we are receiving an event from the _emails_ stream, with ID 1001 and size 15 bytes and then, in the following line the event. We can also send a request like this:\n\n```\nSUBSCRIBE emails 1\n```\n\nIndicating we are going to subscribe to the _emails_ stream from ID 1.\n\nWe will see these commands from/to the server.\n\n&gt; **Warning**\n&gt; The connection must be pipelined. It means that a connection must be attending each request one by one. This should guarantee that the kind of response you are awaiting from the server is the response you need.\n\n## Concepts and types\n\nWe are going to use the following concepts and types in the protocol:\n\n- **stream name**: The name for the streams should be an alphanumeric name of size between 3 and 20. No symbols or spaces are allowed.\n\n- **event id**: The identifier used for the events. It&#39;s used for each stream. When we push an event it&#39;s marked with a sequential number. These numbers cannot be duplicated and the server is in charge to assigned them to the event when it&#39;s created. The lowest number will be always 1.\n\n- **event**: The content of the event must be a valid JSON, an object that we could filter based on any of the root keys provided.\n\n- **aggregator id**: The identifier used for the aggregator element to be stored. It&#39;s recommended you use UUID but actually, you could use whatever while it&#39;s identifying uniquely to the aggregation element.\n\n- **aggregator name**: The name assigned to the kind or type of aggregators. This name should be an alphanumeric name of size between 3 and 20. No symbols or spaces are allowed.\n\n- **aggregator content**: The information to be stored for the aggregator. It should be a valid JSON object.\n\n## Commands or Requests\n\nThis is the list of commands or requests that could be triggered from the client to the server or vice versa.\n\n### PING\n\nThis could be received from/sent to the server. The interaction could be:\n\n```\nS: PING\nC: PONG\n```\n\nOr otherwise:\n\n```\nC: PING\nS: PONG\n```\n\nThe server could be configured with the _keep-alive_ option and depending on the time, it could be sending a `PING` message every X seconds of inactivity. If the server isn&#39;t receiving the response from the client in a specified time (default 5 seconds) the server closes the connection.\n\nThe same logic could be implemented in the client. We could send a `PING` to the server and if the server isn&#39;t replying in a specific number of seconds, the client should close that connection.\n\n&gt; **Note**\n&gt; It&#39;s intended that the connection should be blocked until we receive the `PONG` response. That&#39;s good idea if we need perform a pause in the communication because we detected a big pressure in the channel.\n\n### EVENT\n\nThe information received from the server. We are receiving events when they are published and stored correctly in the persistent storage. The event is formed using the stream name, the event ID and size, and then the event content as a new line:\n\n```\nS: EVENT emails 1001 15\nS: {&quot;name&quot;:&quot;peter&quot;}\n```\n\n### EVENT SUBSCRIBE\n\nThe subscription is performed only from the client to the server. The subscription is accepting the following parameters:\n\n- `stream` the name to the stream to be subscribed.\n- `event-id` the ID of the event where we want to start to retrieve the events.\n\n&gt; **Warning**\n&gt; Be careful indicating a lower number for a stream very populated because it could block the responses from the server until you receive all of the events required.\n\nThe response to this request is a summary indicating the initial ID for the event we are going to receive and the last ID for the last event we are going to receive and, of course, all of the events, one by one:\n\n```\nC: EVENT SUBSCRIBE emails 1\nS: EVENT SUBSCRIBED emails 1 1001\nS: ...\nS: EVENT emails 1001 15\nS: {&quot;name&quot;:&quot;peter&quot;}\n```\n\nIf there are no events inside of the stream at the moment, we will receive:\n\n```\nC: EVENT SUBSCRIBE removals 1\nS: EVENT SUBSCRIBED emails EMPTY\n```\n\n### EVENT PUBLISH\n\nThe client could request the store of an event to the server. The publication of the event requires the name of the stream where the event is going to be published and the size of the event. Followed to that, the event is sent:\n\n```\nC: EVENT PUBLISH emails 14\nC: {&quot;name&quot;:&quot;tom&quot;}\nS: EVENT PUBLISHED emails 1002\n```\n\nIf the event couldn&#39;t be published then the response will be:\n\n```\nC: EVENT PUBLISH emails 14\nC: {&quot;name&quot;:&quot;peter&quot;}\nS: EVENT NON PUBLISHED emails &quot;Invalid JSON&quot;\n```\n\n### EVENT UNSUBSCRIBE\n\nIt&#39;s removing the subscription for the current connection. It&#39;s performed by the client.\n\n```\nC: EVENT UNSUBSCRIBE emails\nS: EVENT UNSUBSCRIBED emails\n```\n\n&gt; **Note**\n&gt; The server should flush all of the events waiting to be sent to the client when an unsubscribe request is performed.\n\n### EVENT LIST\n\nThe client could request the list of the streams available in the system. The response is a list of the streams:\n\n```\nC: EVENT LIST\nS: EVENT LISTED 2 emails users\n```\n\n### EVENT REMOVE\n\nThe client could request the removal of an event stream completely from the persistent storage:\n\n```\nC: EVENT REMOVE emails 1001\nS: EVENT REMOVED emails 1001\n```\n\n&gt; **Warning**\n&gt; The event removal is removing completely the event stream and all of its content. It&#39;s not possible to recover the information. That&#39;s because we could configure each stream for disabling this action. In this case, we could receive a `EVENT NON REMOVED` response.\n\n### AGGREGATE SET\n\nThe aggregation request is performed by the client and it&#39;s storing the information from an aggregator into the persistent storage. The aggregator should be identified by a name and ID for the aggregator, the size of the aggregator content and then the content: \n\n```\nC: AGGREGATE SET emails 27d1b5a0-c54f-4664-a549-b876b0bb3661 15\nC: {&quot;emails&quot;:1002}\nS: AGGREGATE SET emails 27d1b5a0-c54f-4664-a549-b876b0bb3661\n```\n\nIf there are errors presented the return will be as follows:\n\n```\nC: AGGREGATE SET emails 1 12\nC: {&quot;emails&quot;:1002}\nS: AGGREGATE NON SET emails 1 &quot;Invalid JSON&quot;\n```\n\n### AGGREGATE REMOVE\n\nIt&#39;s similar to `AGGREGATE SET` but it&#39;s not indicating the content, only the name and ID for the aggregation to be deleted. The response is a bit different:\n\n```\nC: AGGREGATE REMOVE emails 27d1b5a0-c54f-4664-a549-b876b0bb3661\nS: AGGREGATE REMOVED emails 27d1b5a0-c54f-4664-a549-b876b0bb3661\n```\n\nThe action is always returning the correct state even when the object to be removed is not found.\n\n### AGGREGATE GET\n\nIt&#39;s retrieving the aggregation from the persistent storage:\n\n```\nC: AGGREGATE GET emails 27d1b5a0-c54f-4664-a549-b876b0bb3661\nS: AGGREGATE GOT emails 27d1b5a0-c54f-4664-a549-b876b0bb3661 15\nS: {&quot;emails&quot;:1002}\n```\n\nIf the aggregate isn&#39;t found the response is as follows:\n\n```\nC: AGGREGATE GET emails 1\nS: AGGREGATE NON GOT emails 1 &quot;Not found&quot;\n```\n\n### AGGREGATE LIST\n\nIt&#39;s retrieving the list of aggregators available in the system:\n\n```\nC: AGGREGATE LIST\nS: AGGREGATE LISTED 5 emails users accounts orders payments\n```\n\nThe return is giving the number of aggregators and the list of them.\n\n### PROJECTION CREATE\n\nThe projection stores information processed after the event and it&#39;s adding the information like with the aggregator but adds more indexes for searching the information based on different fields. Using this request we can create a projection with the following information:\n\n```\nC: PROJECTION CREATE users 6 +name +email position address city *country salary\nS: PROJECTION CREATED users\n```\n\nThe plus (+) signs are indicating the fields are unique keys and the star (*) signs are indicating the fields are indexes. The response is indicating the name of the projection created.\n\n### PROJECTION DROP\n\nIt&#39;s removing the projection from the persistent storage:\n\n```\nC: PROJECTION DROP users\nS: PROJECTION DROPPED users\n```\n\nThe action is always returning the correct state even when the object to be removed is not found.\n\n### PROJECTION SET\n\nThe projection stores the information required in the persistent storage. The request is similar to the aggregator but it&#39;s adding more indexes for searching the information based on different fields. Using this request we can create a projection with the following information:\n\n```\nC: PROJECTION SET users 130\nC: {&quot;name&quot;:&quot;peter&quot;,&quot;email&quot;:&quot;peter@mail.com&quot;,&quot;position&quot;:&quot;developer&quot;,&quot;address&quot;:&quot;street 1&quot;,&quot;city&quot;:&quot;London&quot;,&quot;country&quot;:&quot;UK&quot;,&quot;salary&quot;:1000}\nS: PROJECTION SET users peter\n```\n\nThe content must contain all of the fields indicated in the creation of the projection. The response is indicating the name of the projection created.\n\nIf there are errors presented the return will be as follows:\n\n```\nC: PROJECTION SET users 2\nC: {}\nS: PROJECTION NON SET users &quot;Missing required fields&quot;\n```\n\n### PROJECTION SELECT\n\nIt&#39;s retrieving the projection from the persistent storage based on the expression provided:\n\n```\nC: PROJECTION SELECT users 84\nC: city IN (&quot;London&quot;,&quot;Madrid&quot;) AND position != &quot;commercial&quot; AND email =~ &quot;@mail.com$&quot;\nS: PROJECTION SELECTED users 132\nS: [{&quot;name&quot;:&quot;peter&quot;,&quot;email&quot;:&quot;peter@mail.com&quot;,&quot;position&quot;:&quot;developer&quot;,&quot;address&quot;:&quot;street 1&quot;,&quot;city&quot;:&quot;London&quot;,&quot;country&quot;:&quot;UK&quot;,&quot;salary&quot;:1000}]\n```\n\nThe expression sent to the server is based on the SQL syntax. The response is a list of the projections found. You can use different comparison operators for creating the expression:\n\n- `=`, `!=`, `&gt;`, `&lt;` `&gt;=` `&lt;=`: comparison operators\n- `AND`, `OR`: logical operators\n- `IN`: for checking if the value is inside of a list of values\n- `=~`: for checking if the value is like the regular expression provided\n\n&gt; **Note**\n&gt; The projection should have all of the information you need. If you think you need something like JOINs then you should create a new projection with the information you need.\n\n### PROJECTION UPDATE\n\nIt&#39;s updating the projection from the persistent storage based on the expression provided:\n\n```\nC: PROJECTION UPDATE users 27 22\nC: city IN (&quot;London&quot;,&quot;Madrid&quot;)\nC: {&quot;position&quot;:&quot;manager&quot;}\nS: PROJECTION UPDATED users 1\n```\n\nThe return is indicating the number of projections updated.\n\nIn this case, we are sending two expressions, the first one is for selecting the projections to be updated and the second one is for updating the information. As an advantage, the JSON sent will be evaluated, so we can add expressions like:\n\n```\nC: PROJECTION UPDATE users 18 43\nC: city IN (&quot;London&quot;)\nC: {&quot;position&quot;:&quot;manager&quot;,&quot;salary&quot;:salary+1000}\nS: PROJECTION UPDATED users 1\n```\n\n### PROJECTION DELETE\n\nit&#39;s deleting the projection entries from the persistent storage based on the expression provided:\n\n```\nC: PROJECTION DELETE users 17\nC: city IN (&quot;Miami&quot;)\nS: PROJECTION DELETED users 12\n```\n\nWe are removing the projections based on the expression provided. The return is indicating the number of projections deleted.\n","url":"protocol.html"},{"doc":"ream","title":"ream","content":"# Storage\n\nBecause we have three different ways to handle information we have to design three different ways to store or persist that information.\n\nFor example, the events have to be placed in a fast to append system with a index to localize fast each element based on the incremental ID provided for each event.\n\nWe are going to split these and explain these features for ensuring we are choosing a good method for storing them. But first we are going to talk about the files and how the information is going to be stored, after that we could go to the specifics for each of the ways we are handling information for.\n\n## Storage\n\nThe content inside of the file is going to be in binary format based on the `term_to_binary/1` function from BEAM to store and retrieve that information as fast as possible but with the addition in the beginning of a 48-bits integer indicating the size of the element for helping us to navigate the elements if needed.\n\nThe files shouldn&#39;t be too huge, that&#39;s meaning we are needing to create as many files as possible and placing the information in these files. We need to different ways for storing information:\n\n- raw information should be inside of the file and depending on the size we could configure a cache for keeping some of that information for a while.\n- indexes is better to keep them inside of the memory but we have to talk about them independently because that&#39;s not the same in each element.\n\nThe organization of the files is going to be based on directories. The main idea is to use the UUID format, i.e. `276ea36d-02c9-4ea0-996a-72f4162a9fc3` is split in 5 chunks. When we are creating a file we are going to generate a new UUID and based on the data it will be replacing the `-` for `/`, then the path for that file should be `276ea36d/02c9/4ea0/996a/72f4162a9fc3` where the root directory will be `276ea36d` and the name of the file under the last directory is going to be `72f4162a9fc3`.\n\nThe index for the files is going to store the files opened putting the UUID in its binary format and based on the position inside of the file it means if that&#39;s considered the number 1, 2, 3, and so on.\n\n## Schemas\n\nIndeed we have only schemas for projections and the rest of the systems are only storing information about:\n\n- `name` for the stream of events, collection of similar aggregators or tables for projections.\n- `type` where we could say: `stream`, `aggregator`, or `projection`.\n\nThe files for these elements are going to be created under a directory with that name: `type/name`; i.e. if we are creating the stream `users` then the files for storing the events for that stream will be placed under `stream/users`.\n\nIn the case of the projections we are placing an extra file inside of that directory called `schema`. The file is containing a list of the fields available for the projection the content for each field is:\n\n- `name` of the field.\n- `type` of the filed. It could be one of these: `normal`, `index`, `unique`.\n\nThis way we have all of the information about the information we could find about the projection and the information to generate the indexes.\n\n## Event Sourcing\n\nThe event sourcing has the following actions:\n\n- Append a new event to the corresponding stream.\n- Retrieve a list of events based on the correlative ID.\n\nThat said, it&#39;s clear that if we are not needing to modify that data we could create files where that information could be placed with a different limits, we need to create a new file:\n\n- based on date, or\n- based on the size of the file, or\n- based on the number of elements inside of the file.\n\nThe first one is not a good fit because if one stream isn&#39;t used only on specific days it could mean there could be empty files and very populated files. The second one looks good but it could be happening that the each event could be big enough to overflow the limit and we get a file per event. The third one is a bit risky as well in the same case of the size, if the events are big and the number is too high it could mean the file could be huge and difficult to navigate.\n\nI think the combination of the second of the third could give us a good solution. I mean, we could define a minimum limit for the file and if that limit is reached then one condition is achieved. The second condition should be a minimum of elements in terms of getting that number of elements even if the limit of the file was reached.\n\nWe could implement a much more flexible ways for handling the files but for the first version I think that the configuration for:\n\n- `events.min_items_per_file` using a default value of 1000.\n- `events.min_file_size` using a default value of 100M.\n\nThen we are going to create files based on these parameters, and we are going to be appending that information inside of the files. The index for the event sourcing is going to be another file where we are going to be appending an 64-bit integer where the first 16-bits are going to be used to identified the file and the rest 48-bits are going to be in use to store the size where the event is starting. That said, we just defined some limits:\n\n- The maximum size for an event is 281,474,976,710,656 (2^48) or 256TB. But I don&#39;t recommend to use a size higher than 1MB for an event even less if we are going to trigger thousands of events per second.\n- The maximum number of files kept in the system will be 65,536 (2^16).\n\nIf we are sending events of 1KB based on the default configuration that&#39;s meaning we could store 100,000 events for each file or a total of 6,553,600,000 events using 6TB approx.\n\nThe index for events is going to be useful when we need to retrieve an event based on the correlative ID. Given the number we read that position and retrieve the information about which file we should to ask and the size where that event is starting. The event storage will provide us the size of the event and the content for the event.\n\n## Aggregations\n\nThe aggregators are storing their information based on a key and a value, a document. It&#39;s clear this content is going to be updated and read more than inserted as a new element. However, I think the best approach is a copy-on-write strategy. If we need to modify the aggregator we insert a new element inside of the files in the same way we do with the events and then we update the index.\n\nFor avoiding some race-conditions that could arise reading the aggregator when we are processing an update the process in charge of the modification could lock the possible reading actions until the new element is updated inside of the index. However, there&#39;s situations where an eventual consistency is ok. We could implement a configuration parameter:\n\n- `aggregator.read_lock_when_writing` using a default value of `true`, we encourage the consistency.\n\n&gt; **Note**\n&gt; However, it&#39;s not going to lock the events and it&#39;s not going to lock other aggregators or projections.\n\nThe process file has the information about the elements it&#39;s storing and if all of them are marked as deleted, then the file is removed. Based on that the information of the aggregators could be generated again from the events makes no sense to keep that information.\n\nAs another option, we could add a frequency to perform a data vacuum. That&#39;s going to be the process to get the information still valid from the older files where there&#39;s removed information and we put that information as if that&#39;s was called to be modified. It&#39;s going into the new files and then the old file is removed. We could define:\n\n- `aggregator.vacuum_factor` where we specify a number between 1 and 100 and if that percentage of elements are removed then the file is removed and pushed all of their elements to a new file. The default value is 100.\n- `aggregator.vacuum_frequency` indicate the time in seconds when the vacuum is triggered. The default value is 86400 (one day).\n\n## Projections\n\nThe projections are a bit different because they have indexes a bit more complex and the requests could filter the information to be retrieved based on an expression provided. That&#39;s because each file is going to process their own indexes. This is great because we can parallelize the requests for reading and writing but it makes a bit more difficult the unique indexes.\n\nAbout the indexes we have:\n\n- unique indexes are keeping a record per index across all of the files.\n- indexes are keeping many records per index. There will be one index per file.\n\nThe only one problem is about writing if we have defined unique keys. This is going to have a simplified version out of the files for assigning the unique index or refusing it, we are going to use a Cuckoo filter for ensuring the index wasn&#39;t inside of any of the files.\n\nThe storage of the projections is going to take place as the aggregations and in the same way we will have the vacuum of the information with the following parameters:\n\n- `projector.vacuum_factor` where we specify a number between 1 and 100 and if that percentage of elements are removed then the file is removed and pushed all of their elements to a new file. The default value is 100.\n- `projector.vacuum_frequency` indicate the time in seconds when the vacuum is triggered. The default value is 86400 (one day).\n","url":"storage.html"},{"doc":"ream/storage/file/write","title":"Result","content":"pub type Result {\n  Ok\n  Error(reason: file.Reason)\n}\n\nOk\n\n\nError(reason: file.Reason)\n\n\nSynonyms:\nream/storage/file/write.Result\nream/storage/file/write Result","url":"ream/storage/file/write.html#Result"},{"doc":"ream/storage/file/write","title":"ream/storage/file/write","content":"","url":"ream/storage/file/write.html"},{"doc":"ream/storage/file/read","title":"Result","content":"pub type Result {\n  Ok(data: BitString)\n  Eof\n  Error(reason: file.Reason)\n}\n\nOk(data: BitString)\n\n\nEof\n\n\nError(reason: file.Reason)\n\n\nSynonyms:\nream/storage/file/read.Result\nream/storage/file/read Result","url":"ream/storage/file/read.html#Result"},{"doc":"ream/storage/file/read","title":"ream/storage/file/read","content":"","url":"ream/storage/file/read.html"},{"doc":"ream/storage/file/close","title":"Result","content":"pub type Result {\n  Ok\n  Error(reason: file.Reason)\n}\n\nOk\n\n\nError(reason: file.Reason)\n\n\nSynonyms:\nream/storage/file/close.Result\nream/storage/file/close Result","url":"ream/storage/file/close.html#Result"},{"doc":"ream/storage/file/close","title":"ream/storage/file/close","content":"","url":"ream/storage/file/close.html"},{"doc":"ream/storage/file","title":"EncodingType","content":"pub type EncodingType {\n  Unicode\n  Utf8\n  Utf16(Endian)\n  Utf32(Endian)\n  Latin1\n}\n\nUnicode\n\n\nUtf8\n\n\nUtf16(Endian)\n\n\nUtf32(Endian)\n\n\nLatin1\n\n\nSynonyms:\nream/storage/file.EncodingType\nream/storage/file EncodingType","url":"ream/storage/file.html#EncodingType"},{"doc":"ream/storage/file","title":"Endian","content":"pub type Endian {\n  Big\n  Little\n}\n\nBig\n\n\nLittle\n\n\nSynonyms:\nream/storage/file.Endian\nream/storage/file Endian","url":"ream/storage/file.html#Endian"},{"doc":"ream/storage/file","title":"Location","content":"pub type Location {\n  Bof(Int)\n  Cur(Int)\n  Eof(Int)\n}\n\nBof(Int)\n\n\nCur(Int)\n\n\nEof(Int)\n\n\nSynonyms:\nream/storage/file.Location\nream/storage/file Location","url":"ream/storage/file.html#Location"},{"doc":"ream/storage/file","title":"Mode","content":"pub type Mode {\n  Read\n  Write\n  Append\n  Exclusive\n  Raw\n  Binary\n  DelayedWrite(size: Int, delay: Int)\n  ReadAhead(size: Int)\n  Compressed\n  CompressedOne\n  Encoding(encoding: EncodingType)\n  Ram\n  Sync\n  Directory\n}\n\nRead\n\n\nWrite\n\n\nAppend\n\n\nExclusive\n\n\nRaw\n\n\nBinary\n\n\nDelayedWrite(size: Int, delay: Int)\n\n\nReadAhead(size: Int)\n\n\nCompressed\n\n\nCompressedOne\n\n\nEncoding(encoding: EncodingType)\n\n\nRam\n\n\nSync\n\n\nDirectory\n\n\nSynonyms:\nream/storage/file.Mode\nream/storage/file Mode","url":"ream/storage/file.html#Mode"},{"doc":"ream/storage/file","title":"basename","content":"pub external fn basename(filename: String) -&gt; String\n\nSynonyms:\nream/storage/file.basename\nream/storage/file basename","url":"ream/storage/file.html#basename"},{"doc":"ream/storage/file","title":"close","content":"pub fn close(io_device: Pid) -&gt; Result(Bool, Reason)\n\nSynonyms:\nream/storage/file.close\nream/storage/file close","url":"ream/storage/file.html#close"},{"doc":"ream/storage/file","title":"dirname","content":"pub external fn dirname(filename: String) -&gt; String\n\nSynonyms:\nream/storage/file.dirname\nream/storage/file dirname","url":"ream/storage/file.html#dirname"},{"doc":"ream/storage/file","title":"do_open","content":"pub external fn do_open(\n  filename: String,\n  mode: List(Mode),\n) -&gt; Result(Pid, file.Reason)\n\nSynonyms:\nream/storage/file.do_open\nream/storage/file do_open","url":"ream/storage/file.html#do_open"},{"doc":"ream/storage/file","title":"join","content":"pub external fn join(parts: List(String)) -&gt; String\n\nSynonyms:\nream/storage/file.join\nream/storage/file join","url":"ream/storage/file.html#join"},{"doc":"ream/storage/file","title":"open","content":"pub fn open(filename: String, mode: List(Mode)) -&gt; Result(\n  Pid,\n  Reason,\n)\n\nSynonyms:\nream/storage/file.open\nream/storage/file open","url":"ream/storage/file.html#open"},{"doc":"ream/storage/file","title":"position","content":"pub external fn position(\n  io_device: Pid,\n  location: Location,\n) -&gt; Result(Int, file.Reason)\n\nSynonyms:\nream/storage/file.position\nream/storage/file position","url":"ream/storage/file.html#position"},{"doc":"ream/storage/file","title":"read","content":"pub external fn read(io_device: Pid, bytes: Int) -&gt; read.Result\n\nSynonyms:\nream/storage/file.read\nream/storage/file read","url":"ream/storage/file.html#read"},{"doc":"ream/storage/file","title":"recursive_make_directory","content":"pub fn recursive_make_directory(path: String) -&gt; Result(\n  Bool,\n  Reason,\n)\n\nSynonyms:\nream/storage/file.recursive_make_directory\nream/storage/file recursive_make_directory","url":"ream/storage/file.html#recursive_make_directory"},{"doc":"ream/storage/file","title":"write","content":"pub fn write(io_device: Pid, data: BitString) -&gt; Result(\n  Bool,\n  Reason,\n)\n\nSynonyms:\nream/storage/file.write\nream/storage/file write","url":"ream/storage/file.html#write"},{"doc":"ream/storage/file","title":"ream/storage/file","content":"","url":"ream/storage/file.html"},{"doc":"ream/uuid","title":"from_int","content":"pub fn from_int(uuid: Int) -&gt; BitString\n\nSynonyms:\nream/uuid.from_int\nream/uuid from_int","url":"ream/uuid.html#from_int"},{"doc":"ream/uuid","title":"from_string","content":"pub fn from_string(uuid: String) -&gt; BitString\n\nSynonyms:\nream/uuid.from_string\nream/uuid from_string","url":"ream/uuid.html#from_string"},{"doc":"ream/uuid","title":"new","content":"pub fn new() -&gt; BitString\n\nSynonyms:\nream/uuid.new\nream/uuid new","url":"ream/uuid.html#new"},{"doc":"ream/uuid","title":"parts","content":"pub fn parts(uuid: BitString) -&gt; List(String)\n\nSynonyms:\nream/uuid.parts\nream/uuid parts","url":"ream/uuid.html#parts"},{"doc":"ream/uuid","title":"to_int","content":"pub fn to_int(uuid: BitString) -&gt; Int\n\nSynonyms:\nream/uuid.to_int\nream/uuid to_int","url":"ream/uuid.html#to_int"},{"doc":"ream/uuid","title":"to_string","content":"pub fn to_string(uuid: BitString) -&gt; String\n\nSynonyms:\nream/uuid.to_string\nream/uuid to_string","url":"ream/uuid.html#to_string"},{"doc":"ream/uuid","title":"ream/uuid","content":" This module implements the UUID v4 generation algorithm.","url":"ream/uuid.html"},{"doc":"ream/storage/kv/value","title":"Reason","content":"pub type Reason {\n  CapacityExceeded\n}\n\nCapacityExceeded\n\n\nSynonyms:\nream/storage/kv/value.Reason\nream/storage/kv/value Reason","url":"ream/storage/kv/value.html#Reason"},{"doc":"ream/storage/kv/value","title":"Value","content":"pub type Value {\n  Value(\n    offset: Int,\n    deleted: Bool,\n    data: Option(BitString),\n    file_id: Int,\n  )\n}\n The information for a value. The fields are:\n - `offset`: the offset of the value in the file.\n - `deleted`: if the value is deleted.\n - `data`: the value. It can be `None` if we didn&#39;t read the value yet.\n - `file_id`: the id of the file where the value is stored.\nValue(\n  offset: Int,\n  deleted: Bool,\n  data: Option(BitString),\n  file_id: Int,\n)\n\n\nSynonyms:\nream/storage/kv/value.Value\nream/storage/kv/value Value","url":"ream/storage/kv/value.html#Value"},{"doc":"ream/storage/kv/value","title":"ValueFile","content":"pub type ValueFile {\n  ValueFile(\n    id: Int,\n    handler: Pid,\n    size: Int,\n    max_size: Int,\n    file_path: String,\n  )\n}\n The information for the kv file. The fields are:\n - `id`: the id of the file. It&#39;s intended to be a UUID but it&#39;s stored as an Int.\n - `handler`: the file handler to read and write values.\n - `size`: the size of the file.\n - `max_size`: the maximum size of the file.\n - `file_path`: the path of the file.\nValueFile(\n  id: Int,\n  handler: Pid,\n  size: Int,\n  max_size: Int,\n  file_path: String,\n)\n\n\nSynonyms:\nream/storage/kv/value.ValueFile\nream/storage/kv/value ValueFile","url":"ream/storage/kv/value.html#ValueFile"},{"doc":"ream/storage/kv/value","title":"ValueFileInfo","content":"pub type ValueFileInfo {\n  ValueFileInfo(\n    file_id: Int,\n    size: Int,\n    max_size: Int,\n    entries: Int,\n    deleted: Int,\n  )\n}\n The information for the kv file. The fields are:\n - `file_id`: the id of the file. It&#39;s intended to be a UUID but it&#39;s stored as an Int.\n - `size`: the size of the file.\n - `entries`: the number of entries in the file.\n - `deleted`: the number of deleted entries in the file.\nValueFileInfo(\n  file_id: Int,\n  size: Int,\n  max_size: Int,\n  entries: Int,\n  deleted: Int,\n)\n\n\nSynonyms:\nream/storage/kv/value.ValueFileInfo\nream/storage/kv/value ValueFileInfo","url":"ream/storage/kv/value.html#ValueFileInfo"},{"doc":"ream/storage/kv/value","title":"value_size_bits","content":"pub const value_size_bits: Int = 32\n\nSynonyms:\nream/storage/kv/value.value_size_bits\nream/storage/kv/value value_size_bits","url":"ream/storage/kv/value.html#value_size_bits"},{"doc":"ream/storage/kv/value","title":"close","content":"pub fn close(vfile: ValueFile) -&gt; Result(Nil, Reason)\n Close a kv file. It closes the file handler.\nSynonyms:\nream/storage/kv/value.close\nream/storage/kv/value close","url":"ream/storage/kv/value.html#close"},{"doc":"ream/storage/kv/value","title":"create","content":"pub fn create(base_path: String, max_size: Int) -&gt; Result(\n  ValueFile,\n  Reason,\n)\n Create a new kv file. It creates a new file with a random UUID as the\n path for finding the file.\n\n For example, if the UUID is `f81d4fae-7dec-11d0-a765-00a0c91e6bf6`, the\n file will be created in the following path:\n `base_path/f81d4fae/7dec/11d0/a765/00a0c91e6bf6`.\nSynonyms:\nream/storage/kv/value.create\nream/storage/kv/value create","url":"ream/storage/kv/value.html#create"},{"doc":"ream/storage/kv/value","title":"delete","content":"pub fn delete(vfile: ValueFile, value: Value) -&gt; Result(\n  ValueFile,\n  Reason,\n)\n Delete a value from the kv file. It marks the value as deleted.\n It returns the updated kv file with the new size.\nSynonyms:\nream/storage/kv/value.delete\nream/storage/kv/value delete","url":"ream/storage/kv/value.html#delete"},{"doc":"ream/storage/kv/value","title":"get_file_info","content":"pub fn get_file_info(vfile: ValueFile) -&gt; ValueFileInfo\n\nSynonyms:\nream/storage/kv/value.get_file_info\nream/storage/kv/value get_file_info","url":"ream/storage/kv/value.html#get_file_info"},{"doc":"ream/storage/kv/value","title":"open","content":"pub fn open(path: String, file_id: Int, max_size: Int) -&gt; Result(\n  ValueFile,\n  Reason,\n)\n Open a kv file. It opens the file with the given file id.\n If the file doesn&#39;t exist, it is creating it. The main difference\n with `create` is that `open` doesn&#39;t generate a new UUID.\n It returns the kv file with the file handler and the file size.\nSynonyms:\nream/storage/kv/value.open\nream/storage/kv/value open","url":"ream/storage/kv/value.html#open"},{"doc":"ream/storage/kv/value","title":"read","content":"pub fn read(vfile: ValueFile, offset: Int) -&gt; Result(\n  Value,\n  Reason,\n)\n Read a value from the kv file. It reads the value and returns it as\n a BitString with the following format:\n - 4 bytes: the size of the value\n - 1 byte: 0 if the value is not deleted, 1 if it is\n - n bytes: the value\nSynonyms:\nream/storage/kv/value.read\nream/storage/kv/value read","url":"ream/storage/kv/value.html#read"},{"doc":"ream/storage/kv/value","title":"write","content":"pub fn write(vfile: ValueFile, value: BitString) -&gt; Result(\n  #(ValueFile, Value),\n  Reason,\n)\n\nSynonyms:\nream/storage/kv/value.write\nream/storage/kv/value write","url":"ream/storage/kv/value.html#write"},{"doc":"ream/storage/kv/value","title":"write_value","content":"pub fn write_value(vfile: ValueFile, value: Value) -&gt; Result(\n  ValueFile,\n  Reason,\n)\n Write a value to the kv file. It writes the value in the following\n format:\n - 4 bytes: the size of the value\n - 1 byte: 0 if the value is not deleted, 1 if it is\n - n bytes: the value\n It returns the updated kv file with the new size.\nSynonyms:\nream/storage/kv/value.write_value\nream/storage/kv/value write_value","url":"ream/storage/kv/value.html#write_value"},{"doc":"ream/storage/kv/value","title":"ream/storage/kv/value","content":" Store the information of the values for kv.\n\n The events stored in the file are in the following format:\n - 4 bytes: the size of the event\n - 1 byte: 0 if the event is not deleted, 1 if it is\n - n bytes: the event\n\n The KV system is split into two parts: the key store and the value store.\n The file is related to the value store. As we said it&#39;s a file with the\n content marking the size of the value, if it&#39;s deleted and the value itself.\n\n The file is append only. When a value is written, it&#39;s appended to the end\n of the file. When a value is deleted, it&#39;s not deleted from the file, it&#39;s\n marked as deleted. When a value is updated, it&#39;s deleted and a new value is\n appended to the end of the file.","url":"ream/storage/kv/value.html"},{"doc":"ream/storage/kv/memtable","title":"MemTable","content":"pub type MemTable {\n  MemTable(\n    entries: Map(Int, MemTableEntry),\n    size: Int,\n    max_size: Int,\n  )\n}\n MemTable holds a sorted list of the latest written records. Generally, we\n could implement a sorted list algorithm like skip list, but for simplicity\n we use a map instead.\n\n MemTables have a max capacity and when that is reached, we flush the MemTable\n to disk as a Table (Sorted String Table or SSTable). See the sstable module\n for more details.\n\n The MemTable structure is as follows:\n - `entries`: a map of key hash to MemTableEntry. The key hash is the hash of\n   the key string.\n - `size`: the total size of the MemTable in bytes.\n - `max_size`: the maximum size of the MemTable in bytes.\nMemTable(\n  entries: Map(Int, MemTableEntry),\n  size: Int,\n  max_size: Int,\n)\n\n\nSynonyms:\nream/storage/kv/memtable.MemTable\nream/storage/kv/memtable MemTable","url":"ream/storage/kv/memtable.html#MemTable"},{"doc":"ream/storage/kv/memtable","title":"MemTableEntry","content":"pub type MemTableEntry {\n  MemTableEntry(key: String, value: Value)\n}\n MemTableEntry is a single entry in the MemTable.\nMemTableEntry(key: String, value: Value)\n\n\nSynonyms:\nream/storage/kv/memtable.MemTableEntry\nream/storage/kv/memtable MemTableEntry","url":"ream/storage/kv/memtable.html#MemTableEntry"},{"doc":"ream/storage/kv/memtable","title":"Reason","content":"pub type Reason {\n  CapacityExceeded\n}\n Reason is the reason why a MemTable operation failed.\nCapacityExceeded\n The MemTable is full and cannot accept more entries.\n\nSynonyms:\nream/storage/kv/memtable.Reason\nream/storage/kv/memtable Reason","url":"ream/storage/kv/memtable.html#Reason"},{"doc":"ream/storage/kv/memtable","title":"file_id_size_bytes","content":"pub const file_id_size_bytes: Int = 16\n The file ID size is 128-bit integer corresponding to the UUID\n of the file in binary format.\nSynonyms:\nream/storage/kv/memtable.file_id_size_bytes\nream/storage/kv/memtable file_id_size_bytes","url":"ream/storage/kv/memtable.html#file_id_size_bytes"},{"doc":"ream/storage/kv/memtable","title":"file_offset_size_bytes","content":"pub const file_offset_size_bytes: Int = 4\n The file offset size is 32-bit integer.\nSynonyms:\nream/storage/kv/memtable.file_offset_size_bytes\nream/storage/kv/memtable file_offset_size_bytes","url":"ream/storage/kv/memtable.html#file_offset_size_bytes"},{"doc":"ream/storage/kv/memtable","title":"key_hash_size_bytes","content":"pub const key_hash_size_bytes: Int = 4\n The key hash size is 32-bit integer\nSynonyms:\nream/storage/kv/memtable.key_hash_size_bytes\nream/storage/kv/memtable key_hash_size_bytes","url":"ream/storage/kv/memtable.html#key_hash_size_bytes"},{"doc":"ream/storage/kv/memtable","title":"key_size_bytes","content":"pub const key_size_bytes: Int = 2\n The key string size data is 16-bit integer\nSynonyms:\nream/storage/kv/memtable.key_size_bytes\nream/storage/kv/memtable key_size_bytes","url":"ream/storage/kv/memtable.html#key_size_bytes"},{"doc":"ream/storage/kv/memtable","title":"payload_size_bytes","content":"pub const payload_size_bytes: Int = 26\n The payload size for the memtable entry, it is the sum of the\n key hash size, file ID size, file offset size and key size.\nSynonyms:\nream/storage/kv/memtable.payload_size_bytes\nream/storage/kv/memtable payload_size_bytes","url":"ream/storage/kv/memtable.html#payload_size_bytes"},{"doc":"ream/storage/kv/memtable","title":"bitstring_to_entry","content":"pub fn bitstring_to_entry(bitstring: BitString) -&gt; #(\n  Int,\n  MemTableEntry,\n)\n bitstring_to_entry converts a bitstring to a MemTableEntry. It is mainly used\n when we need to convert the binary representation of a MemTableEntry to a\n MemTableEntry.\nSynonyms:\nream/storage/kv/memtable.bitstring_to_entry\nream/storage/kv/memtable bitstring_to_entry","url":"ream/storage/kv/memtable.html#bitstring_to_entry"},{"doc":"ream/storage/kv/memtable","title":"contains","content":"pub fn contains(mem_table: MemTable, key: String) -&gt; Bool\n contains checks if the MemTable contains the given key.\nSynonyms:\nream/storage/kv/memtable.contains\nream/storage/kv/memtable contains","url":"ream/storage/kv/memtable.html#contains"},{"doc":"ream/storage/kv/memtable","title":"delete","content":"pub fn delete(mem_table: MemTable, key: String) -&gt; MemTable\n deletes the given key from the MemTable. If the key does not exist, it\n returns the original MemTable. Otherwise, it returns the updated MemTable.\nSynonyms:\nream/storage/kv/memtable.delete\nream/storage/kv/memtable delete","url":"ream/storage/kv/memtable.html#delete"},{"doc":"ream/storage/kv/memtable","title":"entry_to_bitstring","content":"pub fn entry_to_bitstring(entry: MemTableEntry) -&gt; BitString\n entry_to_bitstring converts a MemTableEntry to a bitstring. It is mainly used\n when we need to convert the MemTableEntry to its binary representation.\nSynonyms:\nream/storage/kv/memtable.entry_to_bitstring\nream/storage/kv/memtable entry_to_bitstring","url":"ream/storage/kv/memtable.html#entry_to_bitstring"},{"doc":"ream/storage/kv/memtable","title":"from_entries","content":"pub fn from_entries(entries: Map(Int, MemTableEntry), max_size: Int) -&gt; MemTable\n calculate_entries_size calculates the total size of the entries in the\n MemTable. It is mainly used when we load the MemTable from disk.\nSynonyms:\nream/storage/kv/memtable.from_entries\nream/storage/kv/memtable from_entries","url":"ream/storage/kv/memtable.html#from_entries"},{"doc":"ream/storage/kv/memtable","title":"get","content":"pub fn get(mem_table: MemTable, key: String) -&gt; Result(Value, Nil)\n gets the value of the given key from the MemTable. If the key does not\n exist, it returns an error.\nSynonyms:\nream/storage/kv/memtable.get\nream/storage/kv/memtable get","url":"ream/storage/kv/memtable.html#get"},{"doc":"ream/storage/kv/memtable","title":"get_bounds","content":"pub fn get_bounds(mem_table: MemTable) -&gt; #(Int, Int)\n get_bounds returns the lower and higher bounds of the MemTable. If the\n MemTable is empty, it returns `#(0, 0)`.\nSynonyms:\nream/storage/kv/memtable.get_bounds\nream/storage/kv/memtable get_bounds","url":"ream/storage/kv/memtable.html#get_bounds"},{"doc":"ream/storage/kv/memtable","title":"hash","content":"pub external fn hash(key: String) -&gt; Int\n generates a hash for the given key.\nSynonyms:\nream/storage/kv/memtable.hash\nream/storage/kv/memtable hash","url":"ream/storage/kv/memtable.html#hash"},{"doc":"ream/storage/kv/memtable","title":"new","content":"pub fn new(max_size: Int) -&gt; MemTable\n\nSynonyms:\nream/storage/kv/memtable.new\nream/storage/kv/memtable new","url":"ream/storage/kv/memtable.html#new"},{"doc":"ream/storage/kv/memtable","title":"search_pivot","content":"pub fn search_pivot(mem_table: MemTable) -&gt; Int\n search for a pivot in the MemTable. The pivot is the middle key in the\n MemTable.\nSynonyms:\nream/storage/kv/memtable.search_pivot\nream/storage/kv/memtable search_pivot","url":"ream/storage/kv/memtable.html#search_pivot"},{"doc":"ream/storage/kv/memtable","title":"set","content":"pub fn set(mem_table: MemTable, key: String, value: Value) -&gt; Result(\n  MemTable,\n  Reason,\n)\n set sets the given key to the given value in the MemTable. If the MemTable\n reaches its capacity, it returns an error. Otherwise, it returns the updated\n MemTable.\nSynonyms:\nream/storage/kv/memtable.set\nream/storage/kv/memtable set","url":"ream/storage/kv/memtable.html#set"},{"doc":"ream/storage/kv/memtable","title":"split","content":"pub fn split(mem_table: MemTable, pivot: Int) -&gt; #(\n  MemTable,\n  MemTable,\n  Int,\n)\n splits the MemTable into two MemTables. The first MemTable contains\n MemTableEntries with keys less than the pivot. The second MemTable contains\n MemTableEntries with keys greater than or equal to the pivot. The pivot is\n the middle key in the MemTable.\nSynonyms:\nream/storage/kv/memtable.split\nream/storage/kv/memtable split","url":"ream/storage/kv/memtable.html#split"},{"doc":"ream/storage/kv/memtable","title":"ream/storage/kv/memtable","content":" MemTable is an in-memory data structure that holds the latest written\n records. It is a sorted list of records sorted by the key hash. When the\n MemTable reaches its capacity, it is flushed to disk as a Table (Sorted\n String Table or SSTable). See the sstable module for more details.","url":"ream/storage/kv/memtable.html"},{"doc":"ream/storage/kv/sstable","title":"flush","content":"pub fn flush(mem_table: MemTable, path: String) -&gt; Result(\n  Bool,\n  Reason,\n)\n\nSynonyms:\nream/storage/kv/sstable.flush\nream/storage/kv/sstable flush","url":"ream/storage/kv/sstable.html#flush"},{"doc":"ream/storage/kv/sstable","title":"load","content":"pub fn load(path: String, max_size: Int) -&gt; Result(\n  MemTable,\n  Reason,\n)\n\nSynonyms:\nream/storage/kv/sstable.load\nream/storage/kv/sstable load","url":"ream/storage/kv/sstable.html#load"},{"doc":"ream/storage/kv/sstable","title":"ream/storage/kv/sstable","content":"","url":"ream/storage/kv/sstable.html"},{"doc":"ream/storage/kv","title":"KV","content":"pub type KV {\n  KV(\n    base_path: String,\n    name: String,\n    memtable_ranges: Map(Int, MemTableRange),\n    active_value_file: Option(Int),\n    values: Map(Int, ValueFile),\n    memtables_loaded: Int,\n    max_memtables_loaded: Int,\n    max_memtable_size: Int,\n    max_value_size: Int,\n  )\n}\n\nKV(\n  base_path: String,\n  name: String,\n  memtable_ranges: Map(Int, MemTableRange),\n  active_value_file: Option(Int),\n  values: Map(Int, ValueFile),\n  memtables_loaded: Int,\n  max_memtables_loaded: Int,\n  max_memtable_size: Int,\n  max_value_size: Int,\n)\n\n\nSynonyms:\nream/storage/kv.KV\nream/storage/kv KV","url":"ream/storage/kv.html#KV"},{"doc":"ream/storage/kv","title":"KVInfo","content":"pub type KVInfo {\n  KVInfo(\n    base_path: String,\n    name: String,\n    values: Int,\n    values_size_bytes: Int,\n    memtables_total: Int,\n    memtables_loaded: Int,\n    memtables_loaded_size_bytes: Int,\n    max_memtables_loaded: Int,\n    max_memtable_size: Int,\n    max_value_size: Int,\n  )\n}\n\nKVInfo(\n  base_path: String,\n  name: String,\n  values: Int,\n  values_size_bytes: Int,\n  memtables_total: Int,\n  memtables_loaded: Int,\n  memtables_loaded_size_bytes: Int,\n  max_memtables_loaded: Int,\n  max_memtable_size: Int,\n  max_value_size: Int,\n)\n\n\nSynonyms:\nream/storage/kv.KVInfo\nream/storage/kv KVInfo","url":"ream/storage/kv.html#KVInfo"},{"doc":"ream/storage/kv","title":"MemTableRange","content":"pub type MemTableRange {\n  MemTableRange(\n    lower: Int,\n    upper: Int,\n    memtable: Option(MemTable),\n  )\n}\n\nMemTableRange(lower: Int, upper: Int, memtable: Option(MemTable))\n\n\nSynonyms:\nream/storage/kv.MemTableRange\nream/storage/kv MemTableRange","url":"ream/storage/kv.html#MemTableRange"},{"doc":"ream/storage/kv","title":"close","content":"pub fn close(kv: KV) -&gt; Result(Nil, Nil)\n\nSynonyms:\nream/storage/kv.close\nream/storage/kv close","url":"ream/storage/kv.html#close"},{"doc":"ream/storage/kv","title":"flush","content":"pub fn flush(kv: KV) -&gt; Result(Nil, Nil)\n\nSynonyms:\nream/storage/kv.flush\nream/storage/kv flush","url":"ream/storage/kv.html#flush"},{"doc":"ream/storage/kv","title":"get","content":"pub fn get(kv: KV, key: String) -&gt; #(Result(BitString, Nil), KV)\n\nSynonyms:\nream/storage/kv.get\nream/storage/kv get","url":"ream/storage/kv.html#get"},{"doc":"ream/storage/kv","title":"info","content":"pub fn info(kv: KV) -&gt; KVInfo\n\nSynonyms:\nream/storage/kv.info\nream/storage/kv info","url":"ream/storage/kv.html#info"},{"doc":"ream/storage/kv","title":"open","content":"pub fn open(path: String, name: String, max_memtables_loaded: Int, max_memtable_size: Int, max_value_size: Int) -&gt; KV\n\nSynonyms:\nream/storage/kv.open\nream/storage/kv open","url":"ream/storage/kv.html#open"},{"doc":"ream/storage/kv","title":"set","content":"pub fn set(kv: KV, key: String, value: BitString) -&gt; KV\n\nSynonyms:\nream/storage/kv.set\nream/storage/kv set","url":"ream/storage/kv.html#set"},{"doc":"ream/storage/kv","title":"ream/storage/kv","content":"","url":"ream/storage/kv.html"},{"doc":"ream/storage/stream/index","title":"Index","content":"pub type Index {\n  Index(offset: Int, size: Int, file_id: Int)\n}\n\nIndex(offset: Int, size: Int, file_id: Int)\n\n\nSynonyms:\nream/storage/stream/index.Index\nream/storage/stream/index Index","url":"ream/storage/stream/index.html#Index"},{"doc":"ream/storage/stream/index","title":"IndexFile","content":"pub type IndexFile {\n  IndexFile(handler: Pid, size: Int, file_path: String)\n}\n\nIndexFile(handler: Pid, size: Int, file_path: String)\n\n\nSynonyms:\nream/storage/stream/index.IndexFile\nream/storage/stream/index IndexFile","url":"ream/storage/stream/index.html#IndexFile"},{"doc":"ream/storage/stream/index","title":"event_size_bits","content":"pub const event_size_bits: Int = 24\n event size (in bits) is storing letting us store events \n of 2^24 bits (16MB) of data\nSynonyms:\nream/storage/stream/index.event_size_bits\nream/storage/stream/index event_size_bits","url":"ream/storage/stream/index.html#event_size_bits"},{"doc":"ream/storage/stream/index","title":"file_id_size_bits","content":"pub const file_id_size_bits: Int = 128\n the size of the file id in bytes, it&#39;s 16 bytes or 128 bits\n representing a 128 bit integer (UUID)\nSynonyms:\nream/storage/stream/index.file_id_size_bits\nream/storage/stream/index file_id_size_bits","url":"ream/storage/stream/index.html#file_id_size_bits"},{"doc":"ream/storage/stream/index","title":"index_size_bytes","content":"pub const index_size_bytes: Int = 25\n the size of the index entry in bytes, it&#39;s 25 bytes because\n we&#39;re storing the offset (6 bytes), the event size (3 bytes)\n and the file id (16 bytes).\nSynonyms:\nream/storage/stream/index.index_size_bytes\nream/storage/stream/index index_size_bytes","url":"ream/storage/stream/index.html#index_size_bytes"},{"doc":"ream/storage/stream/index","title":"offset_size_bits","content":"pub const offset_size_bits: Int = 48\n the size of the offset in bytes, it&#39;s 6 bytes or 48 bits\n letting us store offsets of 2^48 bytes (256TB) of data\nSynonyms:\nream/storage/stream/index.offset_size_bits\nream/storage/stream/index offset_size_bits","url":"ream/storage/stream/index.html#offset_size_bits"},{"doc":"ream/storage/stream/index","title":"add","content":"pub fn add(index_file: IndexFile, event_size: Int, file_id: Int) -&gt; #(\n  Index,\n  IndexFile,\n)\n\nSynonyms:\nream/storage/stream/index.add\nream/storage/stream/index add","url":"ream/storage/stream/index.html#add"},{"doc":"ream/storage/stream/index","title":"close","content":"pub fn close(index_file: IndexFile) -&gt; Result(Nil, Reason)\n\nSynonyms:\nream/storage/stream/index.close\nream/storage/stream/index close","url":"ream/storage/stream/index.html#close"},{"doc":"ream/storage/stream/index","title":"count","content":"pub fn count(index_file: IndexFile) -&gt; Int\n\nSynonyms:\nream/storage/stream/index.count\nream/storage/stream/index count","url":"ream/storage/stream/index.html#count"},{"doc":"ream/storage/stream/index","title":"get","content":"pub fn get(index_file: IndexFile, index: Int) -&gt; Result(\n  Index,\n  Reason,\n)\n\nSynonyms:\nream/storage/stream/index.get\nream/storage/stream/index get","url":"ream/storage/stream/index.html#get"},{"doc":"ream/storage/stream/index","title":"get_next","content":"pub fn get_next(index_file: IndexFile) -&gt; Result(Index, Reason)\n\nSynonyms:\nream/storage/stream/index.get_next\nream/storage/stream/index get_next","url":"ream/storage/stream/index.html#get_next"},{"doc":"ream/storage/stream/index","title":"open","content":"pub fn open(path: String) -&gt; Result(IndexFile, Reason)\n\nSynonyms:\nream/storage/stream/index.open\nream/storage/stream/index open","url":"ream/storage/stream/index.html#open"},{"doc":"ream/storage/stream/index","title":"set_pos","content":"pub fn set_pos(index_file: IndexFile, index: Int) -&gt; Result(\n  Int,\n  Reason,\n)\n\nSynonyms:\nream/storage/stream/index.set_pos\nream/storage/stream/index set_pos","url":"ream/storage/stream/index.html#set_pos"},{"doc":"ream/storage/stream/index","title":"ream/storage/stream/index","content":"","url":"ream/storage/stream/index.html"},{"doc":"ream/storage/stream/event","title":"Event","content":"pub type Event {\n  Event(offset: Int, data: BitString)\n}\n\nEvent(offset: Int, data: BitString)\n\n\nSynonyms:\nream/storage/stream/event.Event\nream/storage/stream/event Event","url":"ream/storage/stream/event.html#Event"},{"doc":"ream/storage/stream/event","title":"EventFile","content":"pub type EventFile {\n  EventFile(id: Int, handler: Pid, size: Int, file_path: String)\n}\n The information for the stream file. The fields are:\n - `id`: the id of the file. It&#39;s intended to be a UUID but it&#39;s stored as an Int.\n - `handler`: the file handler to read and write events.\n - `size`: the size of the file.\n - `file_path`: the path of the file.\nEventFile(id: Int, handler: Pid, size: Int, file_path: String)\n\n\nSynonyms:\nream/storage/stream/event.EventFile\nream/storage/stream/event EventFile","url":"ream/storage/stream/event.html#EventFile"},{"doc":"ream/storage/stream/event","title":"close","content":"pub fn close(stream_file: EventFile) -&gt; Result(Nil, Reason)\n Close a stream file. It closes the file handler.\nSynonyms:\nream/storage/stream/event.close\nream/storage/stream/event close","url":"ream/storage/stream/event.html#close"},{"doc":"ream/storage/stream/event","title":"create","content":"pub fn create(base_path: String) -&gt; Result(EventFile, Reason)\n Create a new stream file. It creates a new file with a random UUID as the\n path for finding the file.\n\n For example, if the UUID is `f81d4fae-7dec-11d0-a765-00a0c91e6bf6`, the\n file will be created in the following path:\n `base_path/f81d4fae/7dec/11d0/a765/00a0c91e6bf6`.\nSynonyms:\nream/storage/stream/event.create\nream/storage/stream/event create","url":"ream/storage/stream/event.html#create"},{"doc":"ream/storage/stream/event","title":"open","content":"pub fn open(path: String, file_id: Int) -&gt; Result(\n  EventFile,\n  Reason,\n)\n Open a stream file. It opens the file with the given file id.\n If the file doesn&#39;t exist, it is creating it. The main difference\n with `create` is that `open` doesn&#39;t generate a new UUID.\n It returns the stream file with the file handler and the file size.\nSynonyms:\nream/storage/stream/event.open\nream/storage/stream/event open","url":"ream/storage/stream/event.html#open"},{"doc":"ream/storage/stream/event","title":"read","content":"pub fn read(stream_file: EventFile, offset: Int) -&gt; Result(\n  Event,\n  Reason,\n)\n Read an event from the stream file. It reads the event and returns it as\n a BitString with the following format:\n - 4 bytes: the size of the event\n - n bytes: the event\nSynonyms:\nream/storage/stream/event.read\nream/storage/stream/event read","url":"ream/storage/stream/event.html#read"},{"doc":"ream/storage/stream/event","title":"write","content":"pub fn write(stream_file: EventFile, event: Event) -&gt; EventFile\n Write an event to the stream file. It writes the event in the following\n format:\n - 4 bytes: the size of the event\n - n bytes: the event\n It returns the updated stream file with the new size.\nSynonyms:\nream/storage/stream/event.write\nream/storage/stream/event write","url":"ream/storage/stream/event.html#write"},{"doc":"ream/storage/stream/event","title":"ream/storage/stream/event","content":" Store the information of the event that is being streamed. It is used to\n write and read events from the file.\n\n The events stored in the file are in the following format:\n - 4 bytes: the size of the event\n - n bytes: the event","url":"ream/storage/stream/event.html"},{"doc":"ream/storage/stream","title":"Stream","content":"pub type Stream {\n  Stream(\n    name: String,\n    index: IndexFile,\n    active_file: Option(EventFile),\n    files: Map(Int, EventFile),\n    base_path: String,\n  )\n}\n\nStream(\n  name: String,\n  index: IndexFile,\n  active_file: Option(EventFile),\n  files: Map(Int, EventFile),\n  base_path: String,\n)\n\n\nSynonyms:\nream/storage/stream.Stream\nream/storage/stream Stream","url":"ream/storage/stream.html#Stream"},{"doc":"ream/storage/stream","title":"add_event","content":"pub fn add_event(stream: Stream, event_content: BitString) -&gt; Result(\n  Stream,\n  Reason,\n)\n\nSynonyms:\nream/storage/stream.add_event\nream/storage/stream add_event","url":"ream/storage/stream.html#add_event"},{"doc":"ream/storage/stream","title":"close","content":"pub fn close(stream: Stream) -&gt; Result(Nil, Reason)\n\nSynonyms:\nream/storage/stream.close\nream/storage/stream close","url":"ream/storage/stream.html#close"},{"doc":"ream/storage/stream","title":"get_base_path","content":"pub fn get_base_path(path: String, name: String) -&gt; String\n\nSynonyms:\nream/storage/stream.get_base_path\nream/storage/stream get_base_path","url":"ream/storage/stream.html#get_base_path"},{"doc":"ream/storage/stream","title":"get_event","content":"pub fn get_event(stream: Stream, index: Int) -&gt; Result(\n  BitString,\n  Reason,\n)\n\nSynonyms:\nream/storage/stream.get_event\nream/storage/stream get_event","url":"ream/storage/stream.html#get_event"},{"doc":"ream/storage/stream","title":"open","content":"pub fn open(name: String, path path: String) -&gt; Result(\n  Stream,\n  Reason,\n)\n\nSynonyms:\nream/storage/stream.open\nream/storage/stream open","url":"ream/storage/stream.html#open"},{"doc":"ream/storage/stream","title":"ream/storage/stream","content":"","url":"ream/storage/stream.html"},{"doc":"ream","title":"ream","content":"","url":"ream.html"}]);